"""Module containing sklearn compatible Biterm Topic Model."""
from __future__ import annotations

from typing import Union

import numpy as np
import scipy.sparse as spr
import sklearn
from numpy.typing import ArrayLike

from tweetopic._btm import (
    compute_biterm_set,
    corpus_unique_biterms,
    fit_model,
    predict_docs,
)
from tweetopic._doc import init_doc_words
from tweetopic.exceptions import NotFittedException


class BTM(sklearn.base.TransformerMixin, sklearn.base.BaseEstimator):
    """Implementation of the Biterm Topic Model with Gibbs Sampling
    solver.

    Parameters
    ----------
    n_components: int
        Number of topics in the model.
    n_iterations: int, default 100
        Number of iterations furing fitting.
    alpha: float, default 6.0
        Dirichlet prior for topic distribution.
    beta: float, default 0.1
        Dirichlet prior for topic-word distribution.

    Attributes
    ----------
    components_: array of shape (n_components, n_vocab)
        Conditional probabilities of all terms given a topic.
    topic_distribution: array of shape (n_components,)
        Prior probability of each topic.
    n_features_in_: int
        Number of total vocabulary items seen during fitting.
    """

    def __init__(
        self,
        n_components: int,
        n_iterations: int = 100,
        alpha: float = 6.0,
        beta: float = 0.1,
    ):
        self.n_components = n_components
        self.n_iterations = n_iterations
        self.alpha = alpha
        self.beta = beta
        # Not none for typing reasons
        self.components_ = np.array(0)
        self.topic_distribution = None
        self.n_features_in_ = 0

    @property
    def _fitted(self) -> bool:
        """Property describing whether the model is fitted."""
        return self.topic_distribution is not None

    def _check_fitted(self) -> None:
        """Raise exception if the model is not fitted."""
        if not self._fitted:
            raise NotFittedException

    def get_params(self, deep: bool = False) -> dict:
        """Get parameters for this estimator.

        Parameters
        ----------
        deep: bool, default False
            Ignored, exists for sklearn compatibility.

        Returns
        -------
        dict
            Parameter names mapped to their values.

        Note
        ----
        Exists for sklearn compatibility.
        """
        return {
            "n_components": self.n_components,
            "n_iterations": self.n_iterations,
            "alpha": self.alpha,
            "beta": self.beta,
        }

    def set_params(self, **params) -> BTM:
        """Set parameters for this estimator.

        Returns
        -------
        BTM
            Estimator instance

        Note
        ----
        Exists for sklearn compatibility.
        """
        for param, value in params.items():
            self.__setattr__(param, value)
        return self

    def fit(self, X: Union[spr.spmatrix, ArrayLike], y: None = None):
        """Fits the model using Gibbs Sampling. Detailed description of the
        algorithm in Yan et al. (2013).

        Parameters
        ----------
        X: array-like or sparse matrix of shape (n_samples, n_features)
            BOW matrix of corpus.
        y: None
            Ignored, exists for sklearn compatibility.

        Returns
        -------
        BTM
            The fitted model.

        Note
        ----
        fit() works in-place too, the fitted model is returned for convenience.
        """
        # Converting X into sparse array if it isn't one already.
        X = spr.csr_matrix(X)
        _, self.n_features_in_ = X.shape
        # Calculating the number of nonzero elements for each row
        # using the internal properties of CSR matrices.
        max_unique_words = np.max(np.diff(X.indptr))
        print("Extracting biterms.")
        doc_unique_words, doc_unique_word_counts = init_doc_words(
            X.tolil(),
            max_unique_words=max_unique_words,
        )
        biterms = corpus_unique_biterms(
            doc_unique_words, doc_unique_word_counts
        )
        biterm_set = compute_biterm_set(biterms)
        self.topic_distribution, self.components_ = fit_model(
            n_iter=self.n_iterations,
            alpha=self.alpha,
            beta=self.beta,
            n_components=self.n_components,
            n_vocab=self.n_features_in_,
            biterms=biterm_set,
        )
        return self

    # TODO: Something goes terribly wrong here, fix this

    def transform(self, X: Union[spr.spmatrix, ArrayLike]) -> np.ndarray:
        """Predicts probabilities for each document belonging to each
        topic.

        Parameters
        ----------
        X: array-like or sparse matrix of shape (n_samples, n_features)
            Document-term matrix.

        Returns
        -------
        array of shape (n_samples, n_components)
            Probabilities for each document belonging to each cluster.

        Raises
        ------
        NotFittedException
            If the model is not fitted, an exception will be raised
        """
        self._check_fitted()
        # Converting X into sparse array if it isn't one already.
        X = spr.csr_matrix(X)
        # n_samples, _ = X.shape
        sample_max_unique_words = np.max(np.diff(X.indptr))
        doc_unique_words, doc_unique_word_counts = init_doc_words(
            X.tolil(),
            max_unique_words=sample_max_unique_words,
        )
        return predict_docs(
            topic_distribution=self.topic_distribution,  # type: ignore
            topic_word_distribution=self.components_,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
        )

    def fit_transform(
        self,
        X: Union[spr.spmatrix, ArrayLike],
        y: None = None,
    ) -> np.ndarray:
        """Fits the model, then transforms the given data.

        Parameters
        ----------
        X: array-like or sparse matrix of shape (n_samples, n_features)
            Document-term matrix.
        y: None
            Ignored, sklearn compatibility.

        Returns
        -------
        array of shape (n_samples, n_components)
            Probabilities for each document belonging to each cluster.
        """
        return self.fit(X).transform(X)



import numpy as np
from numba import njit


@njit(fastmath=True)
def sample_categorical(pvals: np.ndarray) -> int:
    """Samples from a categorical distribution given its parameters.

    Parameters
    ----------
    pvals: array of shape (n_clusters, )
        Parameters of the categorical distribution.

    Returns
    -------
    int
        Sample.
    """
    # NOTE: This function was needed as numba's implementation
    # of numpy's multinomial sampling function has some floating point shenanigans going on.
    # Rejection sampling with cummulutative probabilities :)
    cum_prob = 0
    u = np.random.uniform(0.0, 1.0)
    for i in range(len(pvals)):
        cum_prob += pvals[i]
        if u < cum_prob:
            return i
    else:
        # This shouldn't ever happen, but floating point errors can
        # cause such behaviour ever so often.
        return 0


@njit
def norm_prob(prob: np.ndarray) -> None:
    """Normalizes probabilities in place.

    Parameters
    ----------
    prob: ndarray
        Improper probability distribution.
    """
    (n,) = prob.shape
    total = np.sum(prob)
    if total == 0:
        prob[:] = 1 / n
        return
    for i in range(n):
        prob[i] = prob[i] / total



"""Module for utility functions for fitting BTMs"""

import random
from typing import Dict, Tuple, TypeVar

import numpy as np
from numba import njit
from tqdm import tqdm

from tweetopic._prob import norm_prob, sample_categorical


@njit
def doc_unique_biterms(
    doc_unique_words: np.ndarray, doc_unique_word_counts: np.ndarray
) -> Dict[Tuple[int, int], int]:
    (n_max_unique_words,) = doc_unique_words.shape
    biterm_counts = dict()
    for i_term in range(n_max_unique_words):
        count_i = doc_unique_word_counts[i_term]
        term_i = doc_unique_words[i_term]
        if not count_i:
            break
        for j_term in range(n_max_unique_words):
            count_j = doc_unique_word_counts[j_term]
            term_j = doc_unique_words[j_term]
            if not count_j:
                break
            biterm = min(term_i, term_j), max(term_i, term_j)
            if biterm in biterm_counts:
                continue
            if term_j == term_i:
                count = count_i - 1
                if count:
                    biterm_counts[biterm] = count
            else:
                biterm_counts[biterm] = count_i * count_j
    return biterm_counts


T = TypeVar("T")


@njit
def nb_add_counter(dest: Dict[T, int], source: Dict[T, int]):
    """Adds one counter dict to another in place with Numba"""
    for key in source:
        if key in dest:
            dest[key] += source[key]
        else:
            dest[key] = source[key]


@njit
def corpus_unique_biterms(
    doc_unique_words: np.ndarray, doc_unique_word_counts: np.ndarray
) -> Dict[Tuple[int, int], int]:
    n_documents, _ = doc_unique_words.shape
    biterm_counts = doc_unique_biterms(
        doc_unique_words[0], doc_unique_word_counts[0]
    )
    for i_doc in range(1, n_documents):
        doc_unique_words_i = doc_unique_words[i_doc]
        doc_unique_word_counts_i = doc_unique_word_counts[i_doc]
        doc_biterms = doc_unique_biterms(
            doc_unique_words_i, doc_unique_word_counts_i
        )
        nb_add_counter(biterm_counts, doc_biterms)
    return biterm_counts


@njit
def compute_biterm_set(
    biterm_counts: Dict[Tuple[int, int], int]
) -> np.ndarray:
    return np.array(list(biterm_counts.keys()))


@njit
def arrayify_biterms(
    unique_biterms: np.ndarray,
    unique_biterm_counts: np.ndarray,
    biterm_counts: Dict[Tuple[int, int], int],
):
    i_biterm = 0
    for biterm, count in biterm_counts.items():
        w_i, w_j = biterm
        unique_biterms[i_biterm, 0] = w_i
        unique_biterms[i_biterm, 1] = w_j
        unique_biterm_counts[i_biterm] = count
        i_biterm += 1
    unique_biterm_counts[i_biterm] = 0


@njit(fastmath=True)
def add_remove_biterm(
    add: bool,
    i_biterm: int,
    i_topic: int,
    biterms: np.ndarray,
    topic_word_count: np.ndarray,
    topic_biterm_count: np.ndarray,
) -> None:
    change = 1 if add else -1
    topic_word_count[i_topic, biterms[i_biterm, 0]] += change
    topic_word_count[i_topic, biterms[i_biterm, 1]] += change
    topic_biterm_count[i_topic] += change


@njit(fastmath=True)
def add_biterm(
    i_biterm: int,
    i_topic: int,
    biterms: np.ndarray,
    topic_word_count: np.ndarray,
    topic_biterm_count: np.ndarray,
) -> None:
    add_remove_biterm(
        True, i_biterm, i_topic, biterms, topic_word_count, topic_biterm_count
    )


@njit(fastmath=True)
def remove_biterm(
    i_biterm: int,
    i_topic: int,
    biterms: np.ndarray,
    topic_word_count: np.ndarray,
    topic_biterm_count: np.ndarray,
) -> None:
    add_remove_biterm(
        False, i_biterm, i_topic, biterms, topic_word_count, topic_biterm_count
    )


@njit(fastmath=True)
def init_components(
    n_components: int,
    n_vocab: int,
    biterms: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    n_biterms, _ = biterms.shape
    topic_word_count = np.zeros((n_components, n_vocab), dtype=np.uint)
    topic_biterm_count = np.zeros(n_components, dtype=np.uint)
    biterm_topic_assignments = np.empty(n_biterms, dtype=np.uint)
    for i_biterm in range(n_biterms):
        i_topic = random.randint(0, n_components - 1)
        biterm_topic_assignments[i_biterm] = i_topic
        add_biterm(
            i_biterm, i_topic, biterms, topic_word_count, topic_biterm_count
        )
    return biterm_topic_assignments, topic_word_count, topic_biterm_count


@njit(fastmath=True)
def _topic_biterm_probability(
    i_topic: int,
    i_biterm: int,
    alpha: float,
    beta: float,
    n_vocab: int,
    biterms: np.ndarray,
    topic_word_count: np.ndarray,
    topic_biterm_count: np.ndarray,
) -> float:
    term_i = biterms[i_biterm, 0]
    term_j = biterms[i_biterm, 1]
    numerator = (
        (topic_biterm_count[i_topic] + alpha)
        * (topic_word_count[i_topic, term_i] + beta)
        * (topic_word_count[i_topic, term_j] + beta)
    )
    n_topic_words = topic_biterm_count[i_topic] * 2
    denominator = (n_topic_words + n_vocab * beta) ** 2
    return numerator / denominator


@njit(fastmath=True)
def propose_topic_biterm(
    prediction: np.ndarray,
    i_biterm: int,
    n_components: int,
    alpha: float,
    beta: float,
    n_vocab: int,
    biterms: np.ndarray,
    topic_word_count: np.ndarray,
    topic_biterm_count: np.ndarray,
) -> None:
    for i_topic in range(n_components):
        prediction[i_topic] = _topic_biterm_probability(
            i_topic=i_topic,
            i_biterm=i_biterm,
            alpha=alpha,
            beta=beta,
            n_vocab=n_vocab,
            biterms=biterms,
            topic_word_count=topic_word_count,
            topic_biterm_count=topic_biterm_count,
        )
    norm_prob(prediction)


@njit(fastmath=True)
def estimate_parameters(
    alpha: float,
    beta: float,
    n_components: int,
    n_vocab: int,
    n_biterms: int,
    topic_word_count: np.ndarray,
    topic_biterm_count: np.ndarray,
):
    topic_word_distribution = np.empty((n_components, n_vocab))
    topic_distribution = np.empty(n_components)
    # Equation 6
    for i_topic in range(n_components):
        topic_distribution[i_topic] = (topic_biterm_count[i_topic] + alpha) / (
            n_biterms + n_components * alpha
        )
    # Equation 5
    for i_topic in range(n_components):
        n_topic_terms = topic_biterm_count[i_topic] * 2
        for i_term in range(n_vocab):
            topic_word_distribution[i_topic, i_term] = (
                topic_word_count[i_topic, i_term] + beta
            ) / (n_topic_terms + n_vocab * beta)
    return topic_distribution, topic_word_distribution


@njit(fastmath=True)
def _sampling_step(
    alpha: float,
    beta: float,
    n_components: int,
    n_vocab: int,
    biterms: np.ndarray,
    n_biterms: int,
    biterm_topic_assignments: np.ndarray,
    topic_word_count: np.ndarray,
    topic_biterm_count: np.ndarray,
    prediction: np.ndarray,
):
    for i_biterm in range(n_biterms):
        prev_topic = biterm_topic_assignments[i_biterm]
        remove_biterm(
            i_biterm=i_biterm,
            i_topic=prev_topic,
            biterms=biterms,
            topic_word_count=topic_word_count,
            topic_biterm_count=topic_biterm_count,
        )
        propose_topic_biterm(
            prediction=prediction,
            i_biterm=i_biterm,
            n_components=n_components,
            alpha=alpha,
            beta=beta,
            n_vocab=n_vocab,
            biterms=biterms,
            topic_word_count=topic_word_count,
            topic_biterm_count=topic_biterm_count,
        )
        next_topic = sample_categorical(prediction)
        add_biterm(
            i_biterm=i_biterm,
            i_topic=next_topic,
            biterms=biterms,
            topic_word_count=topic_word_count,
            topic_biterm_count=topic_biterm_count,
        )
        biterm_topic_assignments[i_biterm] = next_topic


def fit_model(
    n_iter: int,
    alpha: float,
    beta: float,
    n_components: int,
    n_vocab: int,
    biterms: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    (
        biterm_topic_assignments,
        topic_word_count,
        topic_biterm_count,
    ) = init_components(n_components, n_vocab, biterms)
    n_biterms, _ = biterms.shape
    prediction = np.zeros(n_components)
    iterator = tqdm(range(n_iter), desc="Sampling")
    for _ in iterator:
        _sampling_step(
            alpha=alpha,
            beta=beta,
            n_components=n_components,
            n_vocab=n_vocab,
            biterms=biterms,
            n_biterms=n_biterms,
            biterm_topic_assignments=biterm_topic_assignments,
            topic_word_count=topic_word_count,
            topic_biterm_count=topic_biterm_count,
            prediction=prediction,
        )
    topic_distribution, topic_word_distribution = estimate_parameters(
        alpha=alpha,
        beta=beta,
        n_components=n_components,
        n_vocab=n_vocab,
        n_biterms=n_biterms,
        topic_word_count=topic_word_count,
        topic_biterm_count=topic_biterm_count,
    )
    return topic_distribution, topic_word_distribution


#
# @njit(fastmath=True)
# def fit_model(
#     n_iter: int,
#     alpha: float,
#     beta: float,
#     n_components: int,
#     n_vocab: int,
#     biterms: np.ndarray,
# ) -> Tuple[np.ndarray, np.ndarray]:
#     (
#         biterm_topic_assignments,
#         topic_word_count,
#         topic_biterm_count,
#     ) = init_components(n_components, n_vocab, biterms)
#     n_biterms, _ = biterms.shape
#     prediction = np.zeros(n_components)
#     for i_iter in range(n_iter):
#         n_transferred = 0
#         for i_biterm in range(n_biterms):
#             prev_topic = biterm_topic_assignments[i_biterm]
#             remove_biterm(
#                 i_biterm=i_biterm,
#                 i_topic=prev_topic,
#                 biterms=biterms,
#                 topic_word_count=topic_word_count,
#                 topic_biterm_count=topic_biterm_count,
#             )
#             propose_topic_biterm(
#                 prediction=prediction,
#                 i_biterm=i_biterm,
#                 n_components=n_components,
#                 alpha=alpha,
#                 beta=beta,
#                 n_vocab=n_vocab,
#                 biterms=biterms,
#                 topic_word_count=topic_word_count,
#                 topic_biterm_count=topic_biterm_count,
#             )
#             next_topic = sample_categorical(prediction)
#             add_biterm(
#                 i_biterm=i_biterm,
#                 i_topic=next_topic,
#                 biterms=biterms,
#                 topic_word_count=topic_word_count,
#                 topic_biterm_count=topic_biterm_count,
#             )
#             biterm_topic_assignments[i_biterm] = next_topic
#             # NOTE: Consider branchless, if it affects performance
#             # n_tranferred += int(next_topic != prev_topic)
#             if next_topic != prev_topic:
#                 n_transferred += 1
#         print(f" Iteration {i_iter}: transferred {n_transferred} biterms.")
#     topic_distribution, topic_word_distribution = estimate_parameters(
#         alpha=alpha,
#         beta=beta,
#         n_components=n_components,
#         n_vocab=n_vocab,
#         n_biterms=n_biterms,
#         topic_word_count=topic_word_count,
#         topic_biterm_count=topic_biterm_count,
#     )
#     return topic_distribution, topic_word_distribution
#


@njit(fastmath=True)
def prob_topic_given_biterm(
    w_i: int,
    w_j: int,
    prediction: np.ndarray,
    topic_distribution: np.ndarray,
    topic_word_distribution: np.ndarray,
):
    """Predicts probabilities of topics given biterms."""
    n_topics = topic_distribution.shape[0]
    for i_topic in range(n_topics):
        prediction[i_topic] = (
            topic_distribution[i_topic]
            * topic_word_distribution[i_topic, w_i]
            * topic_word_distribution[i_topic, w_j]
        )
    norm_prob(prediction)


@njit(fastmath=True)
def prob_topic_given_document(
    prediction: np.ndarray,
    doc_biterms: Dict[Tuple[int, int], int],
    topic_distribution: np.ndarray,
    topic_word_distribution: np.ndarray,
):
    """Predicts probabilities of all topics given biterms in a document."""
    n_topics = prediction.shape[0]
    total_biterms = 0
    for biterm_count in doc_biterms.values():
        total_biterms += biterm_count
    p_topic_given_biterm = np.empty(n_topics)
    for i_topic in range(n_topics):
        prediction[i_topic] = 0
        for biterm, biterm_count in doc_biterms.items():
            w_i, w_j = biterm
            prob_topic_given_biterm(
                w_i,
                w_j,
                p_topic_given_biterm,
                topic_distribution,
                topic_word_distribution,
            )
            p_biterm_given_document = biterm_count / total_biterms
            prediction[i_topic] += (
                p_topic_given_biterm[i_topic] * p_biterm_given_document
            )
    norm_prob(prediction)


# TODO: Something wrong with this
@njit(fastmath=True)
def predict_docs(
    topic_distribution: np.ndarray,
    topic_word_distribution: np.ndarray,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
) -> np.ndarray:
    n_docs = doc_unique_words.shape[0]
    n_topics = topic_distribution.shape[0]
    predictions = np.empty((n_docs, n_topics))
    pred = np.empty(n_topics)
    for i_doc in range(n_docs):
        words, word_counts = (
            doc_unique_words[i_doc],
            doc_unique_word_counts[i_doc],
        )
        biterms = doc_unique_biterms(words, word_counts)
        prob_topic_given_document(
            pred, biterms, topic_distribution, topic_word_distribution
        )
        predictions[i_doc, :] = pred
    return predictions


# @njit(fastmath=True)
# def prob_topic_biterm(
#     i_topic: int,
#     biterm: np.ndarray,
#     n_components: int,
#     topic_distribution: np.ndarray,
#     topic_word_distribution: np.ndarray,
# ) -> float:
#     term_i, term_j = biterm
#     evidence = 0
#     for j_topic in range(n_components):
#         # Calculating this everytime is mildly stupid,
#         # should probably come up with sth smarter.
#         evidence += (
#             topic_distribution[j_topic]
#             * topic_word_distribution[j_topic, term_i]
#             * topic_word_distribution[j_topic, term_j]
#         )
#     prior = topic_distribution[i_topic]
#     posterior = (
#         topic_word_distribution[i_topic, term_i]
#         * topic_word_distribution[i_topic, term_j]
#     )
#     if evidence:
#         return prior * posterior / evidence
#     else:
#         return 0
#
#
# @njit(fastmath=True)
# def predict_doc(
#     n_components: int,
#     topic_distribution: np.ndarray,
#     topic_word_distribution: np.ndarray,
#     doc_unique_biterms: np.ndarray,
#     doc_unique_biterm_counts: np.ndarray,
# ) -> np.ndarray:
#     prediction = np.zeros(n_components)
#     n_biterms = np.sum(doc_unique_biterm_counts)
#     n_max_unique_biterms, _ = doc_unique_biterms.shape
#     for i_topic in range(n_components):
#         i_biterm = 0
#         while (
#             i_biterm < n_max_unique_biterms
#             and doc_unique_biterm_counts[i_biterm]
#         ):
#             biterm = doc_unique_biterms[i_biterm]
#             if n_biterms:
#                 prob_biterm_document = (
#                     doc_unique_biterm_counts[i_biterm] / n_biterms
#                 )
#             else:
#                 prob_biterm_document = 0
#             prediction += (
#                 prob_topic_biterm(
#                     i_topic,
#                     biterm,
#                     n_components,
#                     topic_distribution,
#                     topic_word_distribution,
#                 )
#                 * prob_biterm_document
#             )
#             i_biterm += 1
#     return prediction
#



from typing import Tuple

import numpy as np
import scipy.sparse as spr


def init_doc_words(
    doc_term_matrix: spr.lil_matrix,
    max_unique_words: int,
) -> Tuple[np.ndarray, np.ndarray]:
    n_docs, _ = doc_term_matrix.shape
    doc_unique_words = np.zeros((n_docs, max_unique_words)).astype(np.uint32)
    doc_unique_word_counts = np.zeros((n_docs, max_unique_words)).astype(
        np.uint32
    )
    for i_doc in range(n_docs):
        unique_words = doc_term_matrix[i_doc].rows[0]  # type: ignore
        unique_word_counts = doc_term_matrix[i_doc].data[0]  # type: ignore
        for i_unique in range(len(unique_words)):
            doc_unique_words[i_doc, i_unique] = unique_words[i_unique]
            doc_unique_word_counts[i_doc, i_unique] = unique_word_counts[
                i_unique
            ]
    return doc_unique_words, doc_unique_word_counts



"""Module containing tools for fitting a Dirichlet Multinomial Mixture Model."""
from __future__ import annotations

from math import exp, log

import numpy as np
from numba import njit
from tqdm import tqdm

from tweetopic._prob import norm_prob, sample_categorical


@njit
def _remove_add_doc(
    i_doc: int,
    i_cluster: int,
    remove: bool,
    cluster_word_distribution: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_doc_count: np.ndarray,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    max_unique_words: int,
) -> None:
    if remove:
        cluster_doc_count[i_cluster] -= 1
    else:
        cluster_doc_count[i_cluster] += 1
    for i_unique in range(max_unique_words):
        i_word = doc_unique_words[i_doc, i_unique]
        count = doc_unique_word_counts[i_doc, i_unique]
        if not count:
            # Break out when the word is not present in the document
            break
        if remove:
            cluster_word_count[i_cluster] -= count
            cluster_word_distribution[i_cluster, i_word] -= count
        else:
            cluster_word_count[i_cluster] += count
            cluster_word_distribution[i_cluster, i_word] += count


@njit
def remove_doc(
    i_doc: int,
    i_cluster: int,
    cluster_word_distribution: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_doc_count: np.ndarray,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    max_unique_words: int,
) -> None:
    return _remove_add_doc(
        i_doc,
        i_cluster,
        remove=True,
        cluster_word_distribution=cluster_word_distribution,
        cluster_word_count=cluster_word_count,
        cluster_doc_count=cluster_doc_count,
        doc_unique_words=doc_unique_words,
        doc_unique_word_counts=doc_unique_word_counts,
        max_unique_words=max_unique_words,
    )


@njit
def add_doc(
    i_doc: int,
    i_cluster: int,
    cluster_word_distribution: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_doc_count: np.ndarray,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    max_unique_words: int,
) -> None:
    return _remove_add_doc(
        i_doc,
        i_cluster,
        remove=False,
        cluster_word_distribution=cluster_word_distribution,
        cluster_word_count=cluster_word_count,
        cluster_doc_count=cluster_doc_count,
        doc_unique_words=doc_unique_words,
        doc_unique_word_counts=doc_unique_word_counts,
        max_unique_words=max_unique_words,
    )


@njit
def init_clusters(
    cluster_word_distribution: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_doc_count: np.ndarray,
    doc_clusters: np.ndarray,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    max_unique_words: int,
) -> None:
    """Randomly initializes clusters in the model.

    Parameters
    ----------
    cluster_word_count(OUT): array of shape (n_clusters,)
        Contains the amount of words there are in each cluster.
    cluster_word_distribution(OUT): matrix of shape (n_clusters, n_vocab)
        Contains the amount a word occurs in a certain cluster.
    cluster_doc_count(OUT): array of shape (n_clusters,)
        Array containing how many documents there are in each cluster.
    doc_clusters: array of shape (n_docs)
        Contains a cluster label for each document, that has
        to be assigned.
    doc_unique_words: matrix of shape (n_documents, MAX_UNIQUE_WORDS)
        Matrix containing all indices of unique words in the document.
    doc_unique_word_counts: matrix of shape (n_documents, MAX_UNIQUE_WORDS)
        Matrix containing all counts for each unique word in the document.
    max_unique_words: int
        Maximum number of unique words in any document.

    NOTE
    ----
    Beware that the function modifies a numpy array, that's passed in as
    an input parameter. Should not be used in parallel, as race conditions
    might arise.
    """
    n_docs, _ = doc_unique_words.shape
    for i_doc in range(n_docs):
        i_cluster = doc_clusters[i_doc]
        add_doc(
            i_doc=i_doc,
            i_cluster=i_cluster,
            cluster_word_distribution=cluster_word_distribution,
            cluster_word_count=cluster_word_count,
            cluster_doc_count=cluster_doc_count,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
            max_unique_words=max_unique_words,
        )


@njit(fastmath=True)
def _cond_prob(
    i_cluster: int,
    i_document: int,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    n_words: int,
    alpha: float,
    beta: float,
    n_clusters: int,
    n_vocab: int,
    n_docs: int,
    cluster_doc_count: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_word_distribution: np.ndarray,
    max_unique_words: int,
) -> float:
    """Computes the conditional probability of a certain document joining the
    given mixture component.

    Implements formula no. 4 from Yin & Wang (2014).

    Parameters
    ----------
    i_cluster: int
        The label of the cluster.
    i_document: int
        Index of the document in the corpus.
    doc_unique_words: matrix of shape (n_documents, MAX_UNIQUE_WORDS)
        Matrix containing all indices of unique words in the document.
    doc_unique_word_counts: matrix of shape (n_documents, MAX_UNIQUE_WORDS)
        Matrix containing all counts for each unique word in the document.
    n_words: int
        Total number of words in the document.
    alpha: float
        Alpha parameter of the model.
    beta: float
        Beta parameter of the model.
    n_clusters: int
        Number of mixture components in the model.
    n_vocab: int
        Number of total vocabulary items.
    n_docs: int
        Total number of documents.
    cluster_doc_count: array of shape (n_clusters,)
        Array containing how many documents there are in each cluster.
    cluster_word_count: array of shape (n_clusters,)
        Contains the amount of words there are in each cluster.
    cluster_word_distribution: matrix of shape (n_clusters, n_vocab)
        Contains the amount a word occurs in a certain cluster.
    max_unique_words: int
        Maximum number of unique words seen in a document.
    """
    # I broke the formula into different pieces so that it's easier to write
    # I could not find a better way to organize it, as I'm not in total command of
    # the algebra going on here :))
    # I use logs instead of computing the products directly,
    # as it would quickly result in numerical overflow.
    log_norm_term = log(
        (cluster_doc_count[i_cluster] + alpha)
        / (n_docs - 1 + n_clusters * alpha),
    )
    log_numerator = 0
    for i_unique in range(max_unique_words):
        i_word = doc_unique_words[i_document, i_unique]
        count = doc_unique_word_counts[i_document, i_unique]
        if not count:
            # Breaking out at the first word that doesn't occur in the document
            break
        for j in range(count):
            log_numerator += log(
                cluster_word_distribution[i_cluster, i_word] + beta + j,
            )
    log_denominator = 0
    subres = cluster_word_count[i_cluster] + (n_vocab * beta)
    for j in range(n_words):
        log_denominator += log(subres + j)
    res = exp(log_norm_term + log_numerator - log_denominator)
    return res


@njit(parallel=False, fastmath=True)
def predict_doc(
    probabilities: np.ndarray,
    i_document: int,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    n_words: int,
    alpha: float,
    beta: float,
    n_clusters: int,
    n_vocab: int,
    n_docs: int,
    cluster_doc_count: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_word_distribution: np.ndarray,
    max_unique_words: int,
) -> None:
    """Computes the parameters of the multinomial distribution used for
    sampling.

    Parameters
    ----------
    probabilities(OUT): array of shape (n_clusters, )
        Parameters of the categorical distribution.
    i_document: int
        Index of the document in the corpus.
    doc_unique_words: matrix of shape (n_documents, MAX_UNIQUE_WORDS)
        Matrix containing all indices of unique words in the document.
    doc_unique_word_counts: matrix of shape (n_documents, MAX_UNIQUE_WORDS)
        Matrix containing all counts for each unique word in the document.
    n_words: int
        Total number of words in the document.
    alpha: float
        Alpha parameter of the model.
    beta: float
        Beta parameter of the model.
    n_clusters: int
        Number of mixture components in the model.
    n_vocab: int
        Number of total vocabulary items.
    n_docs: int
        Total number of documents.
    cluster_doc_count: array of shape (n_clusters,)
        Array containing how many documents there are in each cluster.
    cluster_word_count: array of shape (n_clusters,)
        Contains the amount of words there are in each cluster.
    cluster_word_distribution: matrix of shape (n_clusters, n_vocab)
        Contains the amount a word occurs in a certain cluster.
    max_unique_words: int
        Maximum number of unique words seen in a document.

    NOTE
    ----
    Beware that the function modifies a numpy array, that's passed in as
    an input parameter. Should not be used in parallel, as race conditions
    might arise.
    """
    # NOTE: we modify the original array here instead of returning a new
    # one, as allocating new arrays in such a nested loop is very inefficient.
    # Obtain all conditional probabilities
    for i_cluster in range(n_clusters):
        probabilities[i_cluster] = _cond_prob(
            i_cluster=i_cluster,
            i_document=i_document,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
            n_words=n_words,
            alpha=alpha,
            beta=beta,
            n_clusters=n_clusters,
            n_vocab=n_vocab,
            n_docs=n_docs,
            cluster_doc_count=cluster_doc_count,
            cluster_word_count=cluster_word_count,
            cluster_word_distribution=cluster_word_distribution,
            max_unique_words=max_unique_words,
        )
    # Normalize probability vector
    norm_prob(probabilities)


@njit(parallel=False)
def _sampling_step(
    alpha: float,
    beta: float,
    n_clusters: int,
    n_vocab: int,
    n_docs: int,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    doc_clusters: np.ndarray,
    cluster_doc_count: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_word_distribution: np.ndarray,
    max_unique_words: int,
    prediction: np.ndarray,
    doc_word_count: np.ndarray,
) -> None:
    for i_doc in range(n_docs):
        # Removing document from previous cluster
        prev_cluster = doc_clusters[i_doc]
        # Removing document from the previous cluster
        remove_doc(
            i_doc,
            prev_cluster,
            cluster_word_distribution=cluster_word_distribution,
            cluster_word_count=cluster_word_count,
            cluster_doc_count=cluster_doc_count,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
            max_unique_words=max_unique_words,
        )
        # Getting new prediction for the document at hand
        predict_doc(
            probabilities=prediction,
            i_document=i_doc,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
            n_words=doc_word_count[i_doc],
            alpha=alpha,
            beta=beta,
            n_clusters=n_clusters,
            n_vocab=n_vocab,
            n_docs=n_docs,
            cluster_doc_count=cluster_doc_count,
            cluster_word_count=cluster_word_count,
            cluster_word_distribution=cluster_word_distribution,
            max_unique_words=max_unique_words,
        )
        new_cluster = sample_categorical(prediction)
        # Adding document back to the newly chosen cluster
        doc_clusters[i_doc] = new_cluster
        add_doc(
            i_doc,
            new_cluster,
            cluster_word_distribution=cluster_word_distribution,
            cluster_word_count=cluster_word_count,
            cluster_doc_count=cluster_doc_count,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
            max_unique_words=max_unique_words,
        )


def fit_model(
    n_iter: int,
    alpha: float,
    beta: float,
    n_clusters: int,
    n_vocab: int,
    n_docs: int,
    doc_unique_words: np.ndarray,
    doc_unique_word_counts: np.ndarray,
    doc_clusters: np.ndarray,
    cluster_doc_count: np.ndarray,
    cluster_word_count: np.ndarray,
    cluster_word_distribution: np.ndarray,
    max_unique_words: int,
) -> None:
    doc_word_count = np.sum(doc_unique_word_counts, axis=1)
    prediction = np.empty(n_clusters)
    iterator = tqdm(range(n_iter), desc="Sampling")
    for _ in iterator:
        _sampling_step(
            alpha=alpha,
            beta=beta,
            n_clusters=n_clusters,
            n_vocab=n_vocab,
            n_docs=n_docs,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
            doc_clusters=doc_clusters,
            cluster_doc_count=cluster_doc_count,
            cluster_word_count=cluster_word_count,
            cluster_word_distribution=cluster_word_distribution,
            max_unique_words=max_unique_words,
            doc_word_count=doc_word_count,
            prediction=prediction,
        )


#
# @njit(parallel=False)
# def fit_model(
#     n_iter: int,
#     alpha: float,
#     beta: float,
#     n_clusters: int,
#     n_vocab: int,
#     n_docs: int,
#     doc_unique_words: np.ndarray,
#     doc_unique_word_counts: np.ndarray,
#     doc_clusters: np.ndarray,
#     cluster_doc_count: np.ndarray,
#     cluster_word_count: np.ndarray,
#     cluster_word_distribution: np.ndarray,
#     max_unique_words: int,
# ) -> None:
#     """Fits the Dirichlet Mixture Model with Gibbs Sampling. Implements
#     algorithm described in Yin & Wang (2014)
#
#     Parameters
#     ----------
#     n_iter: int
#         Number of iterations to conduct.
#     alpha: float
#         Alpha parameter of the model.
#     beta: float
#         Beta parameter of the model.
#     n_clusters: int
#         Number of mixture components in the model.
#     n_vocab: int
#         Number of total vocabulary items.
#     n_docs: int
#         Total number of documents.
#     doc_term_matrix: matrix of shape (n_documents, n_vocab)
#         Contains how many times a term occurs in each document.
#         (Bag of words matrix)
#     doc_unique_words: array of shape (MAX_UNIQUE_WORDS, )
#         Array containing all the ids of unique words in a document.
#     doc_clusters: array of shape (n_docs, )
#         Contains a cluster label for each document, that has
#         to be assigned.
#     doc_unique_words_count: array of shape (n_documents, )
#         Vector containing the number of unique terms in each document.
#     cluster_doc_count(OUT): array of shape (n_clusters,)
#         Array containing how many documents there are in each cluster.
#     cluster_word_count(OUT): array of shape (n_clusters,)
#         Contains the amount of words there are in each cluster.
#     cluster_word_distribution(OUT): matrix of shape (n_clusters, n_vocab)
#         Contains the amount a word occurs in a certain cluster.
#     max_unique_words: int
#         Maximum count of unique words in a document seen in the corpus.
#     """
#     doc_word_count = np.sum(doc_unique_word_counts, axis=1)
#     prediction = np.empty(n_clusters)
#     for iteration in range(n_iter):
#         total_transfers = 0
#         for i_doc in range(n_docs):
#             # Removing document from previous cluster
#             prev_cluster = doc_clusters[i_doc]
#             # Removing document from the previous cluster
#             remove_doc(
#                 i_doc,
#                 prev_cluster,
#                 cluster_word_distribution=cluster_word_distribution,
#                 cluster_word_count=cluster_word_count,
#                 cluster_doc_count=cluster_doc_count,
#                 doc_unique_words=doc_unique_words,
#                 doc_unique_word_counts=doc_unique_word_counts,
#                 max_unique_words=max_unique_words,
#             )
#             # Getting new prediction for the document at hand
#             predict_doc(
#                 probabilities=prediction,
#                 i_document=i_doc,
#                 doc_unique_words=doc_unique_words,
#                 doc_unique_word_counts=doc_unique_word_counts,
#                 n_words=doc_word_count[i_doc],
#                 alpha=alpha,
#                 beta=beta,
#                 n_clusters=n_clusters,
#                 n_vocab=n_vocab,
#                 n_docs=n_docs,
#                 cluster_doc_count=cluster_doc_count,
#                 cluster_word_count=cluster_word_count,
#                 cluster_word_distribution=cluster_word_distribution,
#                 max_unique_words=max_unique_words,
#             )
#             new_cluster = sample_categorical(prediction)
#             if prev_cluster != new_cluster:
#                 total_transfers += 1
#             # Adding document back to the newly chosen cluster
#             doc_clusters[i_doc] = new_cluster
#             add_doc(
#                 i_doc,
#                 new_cluster,
#                 cluster_word_distribution=cluster_word_distribution,
#                 cluster_word_count=cluster_word_count,
#                 cluster_doc_count=cluster_doc_count,
#                 doc_unique_words=doc_unique_words,
#                 doc_unique_word_counts=doc_unique_word_counts,
#                 max_unique_words=max_unique_words,
#             )
#         n_populated = np.count_nonzero(cluster_doc_count)
#         print(
#             f" Iteration {iteration}/{n_iter}: transferred"
#             f" {total_transfers} documents,"
#             f"{n_populated} clusters remain populated.",
#         )



"""Module containing a fully sklearn compatible Dirichlet Mixture Model."""

from __future__ import annotations

from typing import Union

import numpy as np
import scipy.sparse as spr
import sklearn
from numpy.typing import ArrayLike

from tweetopic._dmm import fit_model, init_clusters, predict_doc
from tweetopic._doc import init_doc_words
from tweetopic.exceptions import NotFittedException


class DMM(sklearn.base.TransformerMixin, sklearn.base.BaseEstimator):
    """Implementation of the Dirichlet Mixture Model with Gibbs Sampling
    solver. The class aims to achieve full compatibility with sklearn.

    Parameters
    ----------
    n_components: int
        Number of mixture components in the model.
    n_iterations: int, default 50
        Number of iterations during fitting.
        If you find your results are unsatisfactory, increase this number.
    alpha: float, default 0.1
        Willingness of a document joining an empty cluster.
    beta: float, default 0.1
        Willingness to join clusters, where the terms in the document
        are not present.

    Attributes
    ----------
    components_: array of shape (n_components, n_vocab)
        Describes all components of the topic distribution.
        Contains the amount each word has been assigned to each component
        during fitting.
    cluster_doc_count: array of shape (n_components,)
        Array containing how many documents there are in each cluster.
    n_features_in_: int
        Number of total vocabulary items seen during fitting.
    n_documents: int
        Total number of documents seen during fitting.
    max_unique_words: int
        Maximum number of unique words in a document seen during fitting.
    """

    def __init__(
        self,
        n_components: int,
        n_iterations: int = 50,
        alpha: float = 0.1,
        beta: float = 0.1,
    ):
        self.n_components = n_components
        self.n_iterations = n_iterations
        self.alpha = alpha
        self.beta = beta
        # Not none for typing reasons
        self.components_ = np.array(0)
        self.cluster_doc_count = None
        self.cluster_word_count = None
        self.n_features_in_ = 0
        self.n_documents = 0
        self.max_unique_words = 0

    @property
    def _fitted(self) -> bool:
        """Property describing whether the model is fitted."""
        # If the number of documents seen is more than 0
        # It can be assumed that the model is fitted.
        return bool(self.n_documents)

    def _check_fitted(self) -> None:
        """Raise exception if the model is not fitted."""
        if not self._fitted:
            raise NotFittedException

    def get_params(self, deep: bool = False) -> dict:
        """Get parameters for this estimator.

        Parameters
        ----------
        deep: bool, default False
            Ignored, exists for sklearn compatibility.

        Returns
        -------
        dict
            Parameter names mapped to their values.

        Note
        ----
        Exists for sklearn compatibility.
        """
        return {
            "n_components": self.n_components,
            "n_iterations": self.n_iterations,
            "alpha": self.alpha,
            "beta": self.beta,
        }

    def set_params(self, **params) -> DMM:
        """Set parameters for this estimator.

        Returns
        -------
        DMM
            Estimator instance

        Note
        ----
        Exists for sklearn compatibility.
        """
        for param, value in params:
            self.__setattr__(param, value)
        return self

    def fit(self, X: Union[spr.spmatrix, ArrayLike], y: None = None):
        """Fits the model using Gibbs Sampling. Detailed description of the
        algorithm in Yin and Wang (2014).

        Parameters
        ----------
        X: array-like or sparse matrix of shape (n_samples, n_features)
            BOW matrix of corpus.
        y: None
            Ignored, exists for sklearn compatibility.

        Returns
        -------
        DMM
            The fitted model.

        Note
        ----
        fit() works in-place too, the fitted model is returned for convenience.
        """
        # Converting X into sparse array if it isn't one already.
        X = spr.csr_matrix(X)
        self.n_documents, self.n_features_in_ = X.shape
        # Calculating the number of nonzero elements for each row
        # using the internal properties of CSR matrices.
        print("Initializing components.")
        self.max_unique_words = np.max(np.diff(X.indptr))
        doc_unique_words, doc_unique_word_counts = init_doc_words(
            X.tolil(),
            max_unique_words=self.max_unique_words,
        )
        initial_clusters = np.random.multinomial(
            1,
            np.ones(self.n_components) / self.n_components,
            size=self.n_documents,
        )
        doc_clusters = np.argmax(initial_clusters, axis=1)
        self.cluster_doc_count = np.zeros(self.n_components)
        self.components_ = np.zeros((self.n_components, self.n_features_in_))
        self.cluster_word_count = np.zeros(self.n_components)
        init_clusters(
            cluster_word_distribution=self.components_,
            cluster_word_count=self.cluster_word_count,
            cluster_doc_count=self.cluster_doc_count,
            doc_clusters=doc_clusters,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
            max_unique_words=self.max_unique_words,
        )
        fit_model(
            n_iter=self.n_iterations,
            alpha=self.alpha,
            beta=self.beta,
            n_clusters=self.n_components,
            n_vocab=self.n_features_in_,
            n_docs=self.n_documents,
            doc_unique_words=doc_unique_words,
            doc_unique_word_counts=doc_unique_word_counts,
            doc_clusters=doc_clusters,
            cluster_doc_count=self.cluster_doc_count,
            cluster_word_count=self.cluster_word_count,
            cluster_word_distribution=self.components_,
            max_unique_words=self.max_unique_words,
        )
        return self

    def transform(self, X: Union[spr.spmatrix, ArrayLike]) -> np.ndarray:
        """Predicts probabilities for each document belonging to each
        component.

        Parameters
        ----------
        X: array-like or sparse matrix of shape (n_samples, n_features)
            Document-term matrix.

        Returns
        -------
        array of shape (n_samples, n_components)
            Probabilities for each document belonging to each cluster.

        Raises
        ------
        NotFittedException
            If the model is not fitted, an exception will be raised
        """
        self._check_fitted()
        # Converting X into sparse array if it isn't one already.
        X = spr.csr_matrix(X)
        sample_max_unique_words = np.max(np.diff(X.indptr))
        doc_unique_words, doc_unique_word_counts = init_doc_words(
            X.tolil(),
            max_unique_words=sample_max_unique_words,
        )
        doc_words_count = np.sum(doc_unique_word_counts, axis=1)
        n_docs = X.shape[0]
        predictions = []
        for i_doc in range(n_docs):
            pred = np.zeros(self.n_components)
            predict_doc(
                probabilities=pred,
                i_document=i_doc,
                doc_unique_words=doc_unique_words,
                doc_unique_word_counts=doc_unique_word_counts,
                n_words=doc_words_count[i_doc],
                alpha=self.alpha,
                beta=self.beta,
                n_clusters=self.n_components,
                n_vocab=self.n_features_in_,
                n_docs=n_docs,
                cluster_doc_count=self.cluster_doc_count,  # type: ignore
                cluster_word_count=self.cluster_word_count,  # type: ignore
                cluster_word_distribution=self.components_,  # type: ignore
                max_unique_words=sample_max_unique_words,
            )
            predictions.append(pred)
        return np.stack(predictions)

    def predict_proba(self, X: Union[spr.spmatrix, ArrayLike]) -> np.ndarray:
        """Alias of :meth:`~tweetopic.dmm.DMM.transform` .

        Mainly exists for compatibility with density estimators in
        sklearn.
        """
        return self.transform(X)

    def predict(self, X: Union[spr.spmatrix, ArrayLike]) -> np.ndarray:
        """Predicts cluster labels for a set of documents. Mainly exists for
        compatibility with density estimators in sklearn.

        Parameters
        ----------
        X: array-like or sparse matrix of shape (n_samples, n_features)
            Document-term matrix.

        Returns
        -------
        array of shape (n_samples,)
            Cluster label for each document.

        Raises
        ------
        NotFittedException
            If the model is not fitted, an exception will be raised
        """
        return np.argmax(self.transform(X), axis=1)

    def fit_transform(
        self,
        X: Union[spr.spmatrix, ArrayLike],
        y: None = None,
    ) -> np.ndarray:
        """Fits the model, then transforms the given data.

        Parameters
        ----------
        X: array-like or sparse matrix of shape (n_samples, n_features)
            Document-term matrix.
        y: None
            Ignored, sklearn compatibility.

        Returns
        -------
        array of shape (n_samples, n_components)
            Probabilities for each document belonging to each cluster.
        """
        return self.fit(X).transform(X)
